/*
 * Copyright (C) 2012-2013, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
static const char *pidfile;

static void pidfile_atexit(void)
{
	unlink(pidfile);
	mog_free(pidfile);
}

/*
 * opens a pid file and returns a file descriptor for it
 * mog_pidfile_commit() should be used on the fd returned by
 * this function (often in a separate process)
 * returns < 0 if there is an error and sets errno=EAGAIN
 * if a pid already exists
 *
 *
 * Example: (error checking is left as an exercise to the reader)
 *
 *	pid_t cur_pid;
 *	int fd = mog_pidfile_open("/path/to/pid", &cur_pid);
 *	daemon(0, 0);
 *	mog_pidfile_commit(fd);
 */
int mog_pidfile_open(const char *path, pid_t *cur)
{
	int fd = open(path, O_RDWR|O_CREAT, 0666);
	struct stat sb;
	char buf[sizeof(pid_t) * 8 / 3 + 1];
	ssize_t r;

	if (fd < 0) return fd;
	if (fstat(fd, &sb) < 0) goto err;

	r = pread(fd, buf, sizeof(buf), 0);
	if (r < 0) goto err;
	if (r > 0) {
		char *end;
		long tmp = strtol(buf, &end, 10);

		if (*end == '\n' && tmp > 0 && tmp < INT_MAX) {
			*cur = tmp;
			if (kill(*cur, 0) < 0 && errno == ESRCH) {
				*cur = -1;
				goto out;
			}
			errno = EAGAIN;
			goto err;
		}
	} /* else r == 0 => nothing to kill */

out:
	assert(pidfile == NULL && "already opened pidfile for process");
	pidfile = canonicalize_filename_mode(path, CAN_EXISTING);
	if (!pidfile)
		goto err;

	return fd;
err:
	PRESERVE_ERRNO( close(fd) );
	return -1;
}

/*
 * commits the pidfile pointed to by the given fd
 * and closes the given fd on success.
 * returns -1 on error and sets errno
 * fd should be the return value of mog_pidfile_open();
 */
int mog_pidfile_commit(int fd)
{
	assert(lseek(fd, 0, SEEK_CUR) == 0 && "pidfile offset != 0");
	assert(pidfile && "mog_pidfile_open not called (or unsuccessful)");

	errno = 0;
	if (dprintf(fd, "%d\n", (int)getpid()) <= 1 || errno == ENOSPC) {
		PRESERVE_ERRNO( close(fd) );
		if (errno == ENOSPC)
			PRESERVE_ERRNO( pidfile_atexit() );
		return -1;
	}
	if (close(fd) < 0 && errno != EINTR)
		return -1;

	atexit(pidfile_atexit);

	return 0;
}
